"""Command-line interface options for the readme-ai application."""

from __future__ import annotations

import os
from typing import Optional

import click
from click import Context, Parameter

from readmeai.config.settings import BadgeOptions, ImageOptions


def prompt_for_custom_image(
    context: Optional[Context],
    parameter: Optional[Parameter],
    value: Optional[str],
) -> str:
    """Prompts the user to input a custom image URL.

    Parameters
    ----------
    ctx
        Click context object.
    param
        Click parameter object.
    value
        Value of the parameter.

    Returns
    -------
    str
        Custom image URL.
    """
    if value == ImageOptions.CUSTOM.name:
        return click.prompt("Enter the URL for your custom image logo")
    elif value in ImageOptions.__members__:
        return ImageOptions[value].value
    else:
        raise click.BadParameter(f"Invalid image URL provided: {value}")


align = click.option(
    "-a",
    "--align",
    type=click.Choice(["center", "left"], case_sensitive=False),
    default="center",
    help="Align the text in the README.md file's header to the left or center.",
)
api_key = click.option(
    "-k",
    "--api-key",
    default=os.environ.get("OPENAI_API_KEY", None),
    help="Your GPT language model API key.",
)
badges = click.option(
    "-b",
    "--badges",
    type=click.Choice(
        [opt.value for opt in BadgeOptions], case_sensitive=False
    ),
    default="flat",
    help="""\
        Badge icon style types to select from when generating README.md badges. The following options are currently available:\n
        - flat \n
        - flat-square \n
        - for-the-badge \n
        - plastic \n
        - skills \n
        - skills-light \n
        - social \n
        """,
)
emojis = click.option(
    "-e",
    "--emojis",
    is_flag=True,
    default=False,
    help="This option adds emojis to the README.md file's header sections. For example, the default header for the 'Overview' section generates the markdown code as '## Overview'. Adding the --emojis flag generates the markdown code as '## 📍 Overview'.",
)
image = click.option(
    "-i",
    "--image",
    type=click.Choice(
        [opt.name for opt in ImageOptions], case_sensitive=False
    ),
    default=ImageOptions.BLUE.name,
    callback=prompt_for_custom_image,
    show_choices=True,
    help="""\
        Project logo image displayed in the README file header. The following options are currently available:\n
        - CUSTOM <enter image url> \n
        - BLACK \n
        - BLUE \n
        - GRADIENT \n
        - PURPLE \n
        - YELLOW \n
        """,
)
model = click.option(
    "-m",
    "--model",
    default="gpt-3.5-turbo",
    help="GPT language model to use for generating various sections of the README.md file.",
)
offline = click.option(
    "--offline",
    is_flag=True,
    default=False,
    help="Use this option to generate a README.md file for free without making any LLM API calls. This option generates the same README file structure, and leaves placeholders for content thats generated by the LLM, which include the project slogan inserted in the header section, along with the 'Overview', 'Features', and 'Modules' sections.",
)
output = click.option(
    "-o",
    "--output",
    default="readme-ai.md",
    help="Define the name of the README.md file to generate.",
)
repository = click.option(
    "-r",
    "--repository",
    required=True,
    help="Required option for defining your repository's remote URL link (GitHub, GitLab, BitBucket), or a local directory path to the source code you'd like to document.",
)
temperature = click.option(
    "-t",
    "--temperature",
    default=0.9,
    type=float,
    help="Temperature LLM API parameter for generating README.md file content. A higher temperature value will generate more creative content, while a lower temperature value will generate more predictable content.",
)
max_tokens = click.option(
    "--max-tokens",
    default=3899,
    type=int,
    help="Max tokens LLM API parameter for generating README.md file content. This option defines the maximum number of tokens to generate for each section of the README.md file.",
)
template = click.option(
    "--template",
    help="Template style to use for generating the README.md file.",
)
language = click.option(
    "-l",
    "--language",
    help="Language to generate README.md file content in. The default language is English (en).",
)



"""Data models and functions for configuring the readme-ai CLI tool."""

import os
from enum import Enum
from importlib import resources
from pathlib import Path
from typing import Dict, List, Optional, Union
from urllib.parse import urlparse, urlsplit

import pkg_resources
from pydantic import BaseModel, validator

from readmeai.core.factory import FileHandler
from readmeai.core.logger import Logger

logger = Logger(__name__)


class GitService(str, Enum):
    """
    Enum class for Git service details.
    """

    LOCAL = ("local", None, "{file_path}")
    GITHUB = (
        "github.com",
        "https://api.github.com/repos/",
        "https://github.com/{full_name}/blob/main/{file_path}",
    )
    GITLAB = (
        "gitlab.com",
        "https://api.gitlab.com/v4/projects/",
        "https://gitlab.com/{full_name}/-/blob/master/{file_path}",
    )
    BITBUCKET = (
        "bitbucket.org",
        "https://api.bitbucket.org/2.0/repositories/",
        "https://bitbucket.org/{full_name}/src/master/{file_path}",
    )

    def __new__(cls, host, api_url, file_url) -> object:
        """Create a new instance of the GitService enum."""
        obj = str.__new__(cls, host)
        obj._value_ = host
        obj.host = host
        obj.api_url = api_url
        obj.file_url = file_url
        return obj

    def extract_name_from_host(repo_host: str) -> str:
        """Return the hostname without periods."""
        return repo_host.split(".")[0]


class BadgeOptions(str, Enum):
    """
    Enum for CLI options for README file badge icons.
    """

    FLAT = "flat"
    FLAT_SQUARE = "flat-square"
    FOR_THE_BADGE = "for-the-badge"
    PLASTIC = "plastic"
    SKILLS = "skills"
    SKILLS_LIGHT = "skills-light"
    SOCIAL = "social"


class ImageOptions(str, Enum):
    """
    Enum for CLI options for README file header images.
    """

    CUSTOM = "CUSTOM"
    BLACK = "https://img.icons8.com/external-tal-revivo-regular-tal-revivo/96/external-readme-is-a-easy-to-build-a-developer-hub-that-adapts-to-the-user-logo-regular-tal-revivo.png"
    BLUE = "https://raw.githubusercontent.com/PKief/vscode-material-icon-theme/ec559a9f6bfd399b82bb44393651661b08aaf7ba/icons/folder-markdown-open.svg"
    GRADIENT = "https://img.icons8.com/nolan/96/markdown.png"
    PURPLE = "https://img.icons8.com/external-tal-revivo-duo-tal-revivo/100/external-markdown-a-lightweight-markup-language-with-plain-text-formatting-syntax-logo-duo-tal-revivo.png"
    YELLOW = "https://img.icons8.com/pulsar-color/96/markdown.png"


class CliSettings(BaseModel):
    """CLI options for the readme-ai application."""

    emojis: bool = True
    offline: bool = False


class FileSettings(BaseModel):
    """Pydantic model for configuration file paths."""

    dependency_files: str
    identifiers: str
    ignore_files: str
    language_names: str
    language_setup: str
    output: str
    shields_icons: str
    skill_icons: str


class GitSettings(BaseModel):
    """Codebase repository settings and validations."""

    repository: Union[str, Path]
    source: Optional[str]
    name: Optional[str]

    @validator("repository", pre=True, always=True)
    def validate_repository(cls, value: Union[str, Path]) -> Union[str, Path]:
        """Validate the repository URL or path."""
        if isinstance(value, str):
            path = Path(value)
            if path.is_dir():
                return value
            try:
                parsed_url = urlparse(value)
                if parsed_url.scheme in ["http", "https"] and any(
                    service.host in parsed_url.netloc for service in GitService
                ):
                    return value
            except ValueError:
                pass
        elif isinstance(value, Path) and value.is_dir():
            return value
        raise ValueError(f"Invalid repository URL or path: {value}")

    @validator("source", pre=True, always=True)
    def set_source(cls, value: Optional[str], values: dict) -> str:
        """Sets the Git service source from the repository provided."""
        repo = values.get("repository")
        if isinstance(repo, Path) or (
            isinstance(repo, str) and Path(repo).is_dir()
        ):
            return GitService.LOCAL.host

        parsed_url = urlparse(str(repo))
        for service in GitService:
            if service.host in parsed_url.netloc:
                return service.host
        return GitService.LOCAL.host

    @validator("name", pre=True, always=True)
    def set_name(cls, value: Optional[str], values: dict) -> str:
        """Sets the repository name from the repository provided."""
        repo = values.get("repository")
        if isinstance(repo, Path):
            return repo.name
        elif isinstance(repo, str):
            parsed_url = urlsplit(repo)
            name = parsed_url.path.split("/")[-1]
            return name.removesuffix(".git")
        return "n/a"


class LlmApiSettings(BaseModel):
    """Pydantic model for OpenAI LLM API details."""

    content: str
    endpoint: str
    encoding: str
    model: str
    rate_limit: int
    temperature: float
    tokens: int
    tokens_max: int


class MarkdownSettings(BaseModel):
    """Pydantic model for Markdown code block templates."""

    align: str
    badges_offline: str
    badges_shields: str
    badges_skills: str
    badges_style: str
    contribute: str
    default: str
    features: str
    getting_started: str
    header: str
    header_left: str
    image: str
    modules: str
    modules_widget: str
    overview: str
    slogan: str
    tables: str
    toc: str
    tree: str


class PromptSettings(BaseModel):
    """Pydantic model for OpenAI prompts."""

    features: str
    overview: str
    slogan: str
    summaries: str


class AppConfig(BaseModel):
    """Nested Pydantic model for the entire configuration."""

    cli: CliSettings
    files: FileSettings
    git: GitSettings
    llm: LlmApiSettings
    md: MarkdownSettings
    prompts: PromptSettings


class AppConfigModel(BaseModel):
    """Pydantic model for the entire configuration."""

    app: AppConfig

    class Config:
        """Pydantic model configuration."""

        validate_assignment = True


class ConfigHelper(BaseModel):
    """Helper class to load additional configuration files."""

    conf: AppConfigModel
    dependency_files: List[str] = []
    ignore_files: Dict[str, List[str]] = {}
    language_names: Dict[str, str] = {}
    language_setup: Dict[str, List[str]] = {}

    class Config:
        allow_mutation = True

    def __init__(self, conf: AppConfigModel, **data):
        super().__init__(conf=conf, **data)
        self.load_helper_files()

    def load_helper_files(self):
        """Load helper configuration files."""
        handler = FileHandler()
        conf_path_list = [
            self.conf.app.files.dependency_files,
            self.conf.app.files.ignore_files,
            self.conf.app.files.language_names,
            self.conf.app.files.language_setup,
        ]

        for path in conf_path_list:
            conf_dict = _get_config_dict(handler, path)

            if "dependency_files" in conf_dict:
                self.dependency_files.extend(
                    conf_dict.get("dependency_files", [])
                )
            if "ignore_files" in conf_dict:
                self.ignore_files.update(conf_dict["ignore_files"])
            if "language_names" in conf_dict:
                self.language_names.update(conf_dict["language_names"])
            if "language_setup" in conf_dict:
                self.language_setup.update(conf_dict["language_setup"])


def _get_config_dict(handler: FileHandler, file_path: str) -> dict:
    """Get configuration dictionary from TOML file."""
    try:
        resource_path = resources.files("readmeai.settings") / file_path
        logger.info(f"Resource path using importlib: {resource_path}")
    except TypeError as exc_info:
        logger.debug(f"Error with importlib.resources: {exc_info}")
        try:
            resource_path = Path(
                pkg_resources.resource_filename(
                    "readmeai", f"settings/{file_path}"
                )
            ).resolve()
            logger.info(f"Resource path using pkg_resources: {resource_path}")
        except FileNotFoundError as exc_info:
            logger.debug(f"Error with pkg_resources: {exc_info}")
            raise

    if not os.path.exists(resource_path):
        raise FileNotFoundError(f"Config file not found: {resource_path}")

    return handler.read(resource_path)


def load_config(path: str = "config.toml") -> AppConfig:
    """Load configuration constants from TOML file."""
    handler = FileHandler()
    conf_dict = _get_config_dict(handler, path)
    return AppConfigModel.parse_obj({"app": conf_dict}).app


def load_config_helper(conf: AppConfigModel) -> ConfigHelper:
    """Load multiple configuration helper TOML files."""
    return ConfigHelper(conf=conf)



"""Builds each section of the README Markdown file."""

__package__ = "readmeai"

import re
from typing import List

from readmeai.config.settings import AppConfig, ConfigHelper, GitService
from readmeai.core.factory import FileHandler
from readmeai.markdown import badges, quickstart, tables
from readmeai.services import git_utilities as vcs


def build_readme_md(
    conf: AppConfig,
    helper: ConfigHelper,
    deps: list,
    summaries: tuple,
) -> None:
    """Constructs each section of the README Markdown file."""
    readme_md_sections = format_readme_md(conf, helper, deps, summaries)
    readme_md_file = "\n".join(readme_md_sections)
    FileHandler().write(conf.files.output, readme_md_file)


def format_readme_md(
    conf: AppConfig,
    helper: ConfigHelper,
    deps: list,
    summaries: tuple,
) -> List[str]:
    """Formats the README Markdown file contents for each section."""
    repo_url = conf.git.repository
    user_name, repo_name = vcs.get_remote_full_name(repo_url)
    full_name = f"{user_name}/{repo_name}"

    if conf.git.source == GitService.LOCAL.value:
        repo_path = f"../{repo_name}"
    else:
        repo_path = repo_url

    if "skills" not in conf.md.badges_style:
        md_header = conf.md.header.format(
            alignment=conf.md.align,
            image=conf.md.image,
            repo_name=repo_name.upper(),
            slogan=conf.md.slogan,
        )
        md_badges = badges.shields_icons(conf, deps, full_name)
    else:
        md_header = "<!---->\n"
        md_badges = badges.skill_icons(conf, deps)

    formatted_code_summaries = tables.format_code_summaries(
        conf.md.default,
        summaries,
    )
    md_code_summaries = tables.generate_markdown_tables(
        conf.md.modules_widget, formatted_code_summaries, full_name, repo_url
    )

    md_commands = quickstart.getting_started(conf, helper, deps, summaries)
    md_quick_start = conf.md.getting_started.format(
        repo_name=repo_name,
        repo_url=repo_path,
        install_command=md_commands[0],
        run_command=md_commands[1],
        test_command=md_commands[2],
    )

    md_contribute = conf.md.contribute.format(
        full_name=full_name, repo_name=repo_name
    )

    readme_md_sections = [
        md_header,
        md_badges,
        conf.md.toc.format(repo_name),
        conf.md.overview,
        conf.md.features,
        conf.md.tree,
        conf.md.modules,
        md_code_summaries,
        md_quick_start,
        md_contribute,
    ]

    if conf.cli.emojis is False:
        return remove_emojis_from_headers(readme_md_sections)

    return readme_md_sections


def remove_emojis_from_headers(content_list: List[str]) -> List[str]:
    """Removes emojis from headers and ToC."""
    emoji_pattern = re.compile(
        pattern="["
        "\U0001F600-\U0001F64F"  # emoticons
        "\U0001F300-\U0001F5FF"  # symbols & pictographs
        "\U0001F680-\U0001F6FF"  # transport & map symbols
        "\U0001F700-\U0001F77F"  # alchemical symbols
        "\U0001F780-\U0001F7FF"  # Geometric Shapes Extended
        "\U0001F800-\U0001F8FF"  # Supplemental Arrows-C
        "\U0001F900-\U0001F9FF"  # Supplemental Symbols and Pictographs
        "\U0001FA00-\U0001FA6F"  # Chess Symbols
        "\U0001FA70-\U0001FAFF"  # Symbols and Pictographs Extended-A
        "\U00002702-\U000027B0"  # Dingbats
        "\U000024C2-\U0001F251"  # flags (iOS)
        "]+",
        flags=re.UNICODE,
    )

    # Replace emojis for markdown lines starting with header symbols or in the ToC
    modified_content = []
    for section in content_list:
        lines = section.split("\n")
        for index, line in enumerate(lines):
            if (
                line.startswith("#")
                or "Table of Contents" in section
                or "Quick Links" in section
            ):
                lines[index] = emoji_pattern.sub("", line)
        modified_content.append("\n".join(lines))

    return modified_content



"""GPT language model API handler for document generation."""

import asyncio
import time
import traceback
from contextlib import asynccontextmanager
from typing import Any, Dict, List, Tuple, Union

import openai
from cachetools import TTLCache
from httpx import (
    AsyncClient,
    HTTPStatusError,
    Limits,
    NetworkError,
    TimeoutException,
)
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

import readmeai.config.settings as settings
from readmeai.core.logger import Logger
from readmeai.core.tokens import (
    adjust_max_tokens,
    get_token_count,
    truncate_tokens,
)
from readmeai.utils.utils import format_sentence


class ModelHandler:
    """LLM API handler that generates text for the README.md file."""

    @asynccontextmanager
    async def use_api(self) -> None:
        """Context manager for managing HTTP client resources used by the the LLM API."""
        try:
            yield self
        finally:
            await self.close()

    async def close(self) -> None:
        """Closes the HTTP client."""
        await self.http_client.aclose()

    def __init__(self, config: settings.AppConfig) -> None:
        """Initializes the GPT language model API handler."""
        self.logger = Logger(__name__)
        self.config = config
        self.prompts = config.prompts
        self.tokens = config.llm.tokens
        self.tokens_max = config.llm.tokens_max
        self.cache = TTLCache(maxsize=100, ttl=360)
        self.http_client = AsyncClient(
            http2=True,
            timeout=20,
            limits=Limits(max_keepalive_connections=100, max_connections=200),
        )
        self.last_request_time = time.monotonic()
        self.rate_limit_semaphore = asyncio.Semaphore(config.llm.rate_limit)

    async def batch_text_generator(
        self, prompts: List[Union[str, Tuple[str, str]]]
    ) -> List[str]:
        """Generates text for the README.md file using GPT language models.

        Parameters
        ----------
        prompts
            A list of prompts to generate text from.

        Returns
        -------
        List[str]
            A list of generated text.
        """
        responses = []
        for batch in self._batch_prompts(prompts):
            batch_responses = await asyncio.gather(
                *[self._process_prompt(prompt) for prompt in batch]
            )
            responses.extend(batch_responses)
        return responses

    def _batch_prompts(
        self, prompts: List[Union[str, Tuple[str, str]]], batch_size: int = 5
    ) -> List[List[Union[str, Tuple[str, str]]]]:
        """Batches prompts for the LLM API.

        Parameters
        ----------
        prompts
            A list of prompts to batch.
        batch_size, optional
            The number of prompts to batch, by default 5

        Returns
        -------
        List[List[Union[str, Tuple[str, str]]]]
            A list of batches of prompts.

        Yields
        ------
        List[Union[str, Tuple[str, str]]]
            A batch of prompts.
        """
        for i in range(0, len(prompts), batch_size):
            yield prompts[i : i + batch_size]

    def build_prompt(self, prompt_type, context) -> str:
        """Generates a prompt for the LLM API.

        Parameters
        ----------
        prompt_type
            The type of prompt to generate i.e. features, overview, slogan.
        context
            A dictionary containing the context for the prompt.

        Returns
        -------
        str
            The generated prompt.
        """
        prompt_templates = {
            "features": self.prompts.features,
            "overview": self.prompts.overview,
            "slogan": self.prompts.slogan,
        }
        prompt_template = prompt_templates.get(prompt_type)

        if prompt_template:
            return prompt_template.format(*[context[key] for key in context])
        else:
            self.logger.error(f"Unknown prompt type: {prompt_type}")
            return ""

    async def _process_prompt(self, prompt: Dict[str, Any]) -> str:
        """Processes a prompt and returns the generated text.

        Parameters
        ----------
        prompt
            A dictionary containing the prompt type and context.

        Returns
        -------
        str
            The generated text from the LLM API.
        """
        if prompt["type"] == "summaries":
            return await self._generate_summaries(prompt["context"])
        else:
            formatted_prompt = self.build_prompt(
                prompt["type"], prompt["context"]
            )
            tokens = adjust_max_tokens(self.tokens, formatted_prompt)
            _, summary = await self.generate_text(
                prompt["type"], formatted_prompt, tokens
            )
            return summary

    async def _generate_summaries(
        self, file_context: List[Tuple[str, str]]
    ) -> List[Tuple[str, str]]:
        """Generates code summaries for the README.md file.

        Parameters
        ----------
        file_context
            A list of tuples containing the file path and file content.

        Returns
        -------
        List[Tuple[str, str]]
            List of tuples containing the file path and the generated summary or error message.
        """
        code_summaries = []

        for context in file_context["summaries"]:
            file_path, file_content = context
            prompt = self.prompts.summaries.format(
                self.config.md.tree, file_path, file_content
            )
            tokens = adjust_max_tokens(self.tokens, prompt)
            _, summary_or_error = await self.generate_text(
                file_path, prompt, tokens
            )
            code_summaries.append((file_path, summary_or_error))

        return code_summaries

    @retry(
        stop=stop_after_attempt(4),
        wait=wait_exponential(multiplier=1, min=2, max=6),
        retry=retry_if_exception_type(
            (
                openai.error.OpenAIError,
                HTTPStatusError,
                NetworkError,
                TimeoutException,
            )
        ),
    )
    async def generate_text(
        self, index: str, prompt: str, tokens: int
    ) -> Tuple[str, str]:
        """Generates text for the README.md file using GPT language models.

        Parameters
        ----------
        index
            Type of prompt i.e. features, overview, slogan, or code summary.
        prompt
            The prompt to generate text from.
        tokens
            The number of tokens to generate.

        Returns
        -------
        Tuple[str, str]
            Tuple containing the prompt type and generated response.
        """
        token_count = get_token_count(prompt, self.config.llm.encoding)
        if token_count > self.tokens_max:
            self.logger.warning(
                f"Truncating tokens: {token_count} > {self.tokens_max} max"
            )
            prompt = truncate_tokens(prompt, tokens)

        try:
            async with self.rate_limit_semaphore:
                response = await self.http_client.post(
                    self.config.llm.endpoint,
                    headers={"Authorization": f"Bearer {openai.api_key}"},
                    json={
                        "messages": [
                            {
                                "role": "system",
                                "content": self.config.llm.content,
                            },
                            {"role": "user", "content": prompt},
                        ],
                        "model": self.config.llm.model,
                        "temperature": self.config.llm.temperature,
                        "max_tokens": tokens,
                    },
                )
                response.raise_for_status()
                llm_text = response.json()
                summary = llm_text["choices"][0]["message"]["content"]
                summary = (
                    format_sentence(summary)
                    if index != "features"
                    else summary
                )
                self.cache[prompt] = summary
                self.logger.info(f"Response: {index} - {summary}")
                return index, summary

        except Exception as exc_info:
            error_message = f"Error generating summary: {exc_info.__class__.__name__} occurred. See logs for details."
            self.logger.error(
                f"Exception making request: {exc_info}\n{traceback.format_exc()}"
            )
            return index, error_message



"""Dynamically creates the 'Quick Start' section of the README file."""

import traceback
from pathlib import Path
from typing import List

from readmeai.config.settings import AppConfig, ConfigHelper
from readmeai.core.logger import Logger

logger = Logger(__name__)


def getting_started(
    conf: AppConfig,
    helper: ConfigHelper,
    deps: List[str],
    summaries: List[str],
):
    """Generates the 'Quick Start' section of the README file."""
    install_command = run_command = test_command = conf.md.default

    try:
        language_counts = {}
        for file_path, file_content in summaries:
            language = Path(file_path).suffix[1:]
            if language and language not in helper.ignore_files:
                if language in language_counts:
                    language_counts[language] += 1
                else:
                    language_counts[language] = 1

        if language_counts:
            language_top = max(language_counts, key=language_counts.get)

            if "streamlit" in deps:
                language_top = "streamlit"

            language_name = helper.language_names.get(language_top, "n/a")
            language_setup = helper.language_setup.get(language_name, [])

            logger.info(f"Language counts {language_counts}")
            logger.info(f"Getting started commands: {language_setup}")

            if len(language_setup) >= 3:
                install_command = language_setup[0]
                run_command = language_setup[1]
                test_command = language_setup[2]

    except Exception as exc_info:
        logger.debug(
            f"Exception occurred: {exc_info}\n{traceback.format_exc()}"
        )

    return (install_command, run_command, test_command)



"""Abstract factory for all dependency file parsers."""

from typing import Dict

from readmeai.parsers.base_parser import FileParser
from readmeai.parsers.gomod import GoModParser
from readmeai.parsers.gradle import BuildGradleKtsParser, BuildGradleParser
from readmeai.parsers.maven import MavenParser
from readmeai.parsers.npm import JsonParser
from readmeai.parsers.python import RequirementsParser, TomlParser, YamlParser
from readmeai.parsers.rust import CargoTomlParser


def parser_factory() -> Dict[str, FileParser]:
    """Returns a dictionary of callable file parser methods."""
    return {
        "build.gradle": BuildGradleParser(),
        "build.gradle.kts": BuildGradleKtsParser(),
        "cargo.toml": CargoTomlParser(),
        "go.mod": GoModParser(),
        "package.json": JsonParser(),
        "environment.yml": YamlParser(),
        "environment.yaml": YamlParser(),
        "Pipfile": TomlParser(),
        "pyproject.toml": TomlParser(),
        "requirements.txt": RequirementsParser(),
        "requirements-dev.txt": RequirementsParser(),
        "pom.xml": MavenParser(),
    }
    """
    "yarn.lock": parse_yarn_lock,
    "CMakeLists.txt": parse_cmake,
    "Makefile.am": parse_makefile_am,
    "configure.ac": parse_configure_ac,

    def parse_yarn_lock(content: str) -> List[str]:
        try:
            return re.findall(r"(\S+)(?=@)", content)
        except re.error as excinfo:
            logger.error(f"Error parsing yarn.lock: {str(excinfo)}")
        return []


    def parse_cmake(content: str) -> List[str]:
        regex = re.compile(r"add_executable\([^)]+\s+([^)]+)\)")
        package_names = regex.findall(content)
        return package_names


    def parse_configure_ac(content: str) -> List[str]:
        regex = re.compile(r"AC_CHECK_LIB\([^)]+\s+([^)]+)\)")
        package_names = regex.findall(content)
        return package_names


    def parse_makefile_am(content: str) -> List[str]:
        regex = re.compile(r"bin_PROGRAMS\s*=\s*(.+)")
        package_names = []
        matches = regex.findall(content)
        for match in matches:
            deps = filter(None, match.split())
            package_names.extend(deps)
        return package_names
    """



"""Processes the input codebase for analysis and metadata extraction."""

from pathlib import Path
from typing import Dict, Generator, List, Tuple

from readmeai.config import settings
from readmeai.core.logger import Logger
from readmeai.core.tokens import get_token_count
from readmeai.parsers.factory import parser_factory
from readmeai.utils import utils

logger = Logger(__name__)

PARSERS = parser_factory()


class RepoProcessor:
    """Handles preprocessing of the input codebase."""

    def __init__(
        self,
        config: settings.AppConfig,
        conf_helper: settings.ConfigHelper,
    ):
        self.config = config
        self.config_helper = conf_helper
        self.language_names = conf_helper.language_names
        self.language_setup = conf_helper.language_setup
        self.encoding_name = config.llm.encoding
        self.local_source = settings.GitService.LOCAL.value

    def analyze(self, temp_dir: str) -> List[Dict]:
        """Analyzes a local or remote git repository."""
        contents = self.generate_contents(temp_dir)

        repo_source = self.config.git.source
        if repo_source != self.local_source:
            logger.info(f"Tokenizing content from host: {repo_source}")
            contents = self.tokenize_content(contents)

        contents = self.process_language_mapping(contents)

        return contents

    def generate_contents(self, repo_path: str) -> List[Dict]:
        """Generates a List of Dict of file information."""
        repo_path = Path(repo_path)

        data = list(self.generate_file_info(repo_path))

        contents = []
        for name, path, content in data:
            extension = Path(name).suffix.lstrip(".")
            contents.append(
                {
                    "name": name,
                    "path": path,
                    "content": content,
                    "extension": extension,
                }
            )
        return contents

    def get_dependency_file_contents(self, contents: List[Dict]) -> List[str]:
        """Extracts the dependency file contents using the factory pattern."""
        parsers = PARSERS

        filtered = [c for c in contents if c["name"] in parsers]
        if not filtered:
            return []

        parsed_contents = []
        for content in filtered:
            parser = parsers.get(content["name"])
            parsed_content = parser.parse(content=content["content"])
            parsed_contents.append(parsed_content)
            logger.info(
                f"Dependency file found {content['name']}\n\t{parsed_content}"
            )
        return utils.flatten_list(parsed_contents)

    def parse_content(content: Dict) -> List[str]:
        """Helper function to parse the content of a file."""
        parser = PARSERS.get(content["name"])
        if parser:
            return parser.parse(content["content"])
        return []

    def generate_file_info(
        self, repo_path: Path
    ) -> Generator[Tuple[str, Path, str], None, None]:
        """Generates a tuple of file information."""
        for file_path in repo_path.rglob("*"):
            if utils.should_ignore(self.config_helper, file_path):
                continue

            if file_path.is_file():
                if ".github/workflows" in str(
                    file_path.relative_to(repo_path)
                ):
                    yield "github actions", file_path.relative_to(
                        repo_path
                    ), ""
                try:
                    with file_path.open(encoding="utf-8") as file:
                        content = file.read()
                    relative_path = file_path.relative_to(repo_path)
                    yield file_path.name, relative_path, content
                except UnicodeDecodeError:
                    continue

    def get_file_contents(self, contents: Dict) -> Dict[str, str]:
        """Extracts the file contents from the list of dicts."""
        return {content["path"]: content["content"] for content in contents}

    def get_unique_contents(
        self, contents: Dict, keys: List[str]
    ) -> List[str]:
        """Extracts the unique contents from the list of dicts."""
        unique_contents = {data[key] for key in keys for data in contents}
        return list(unique_contents)

    def get_dependencies(
        self, temp_dir: str = None
    ) -> Tuple[List[str], Dict[str, str]]:
        """Extracts the dependencies of the user's repository."""
        contents = self.analyze(temp_dir)
        dependencies = self.get_dependency_file_contents(contents)
        attributes = ["extension", "language", "name"]
        dependencies.extend(self.get_unique_contents(contents, attributes))
        return list(set(dependencies)), self.get_file_contents(contents)

    def process_language_mapping(self, contents: List[Dict]) -> List[Dict]:
        """Maps file extensions to their programming languages."""
        for content in contents:
            content["language"] = self.language_names.get(
                content["extension"], ""
            ).lower()
            setup = self.language_setup.get(content["language"], "")
            setup = setup if isinstance(setup, list) else [None, None]
            while len(setup) < 3:
                setup.append(None)
            content["install"], content["run"], content["test"] = setup
        return contents

    def tokenize_content(self, contents: List[Dict]) -> List[Dict]:
        """Tokenizes the content of each file."""
        for content in contents:
            content["tokens"] = get_token_count(
                content["content"], self.encoding_name
            )
        return contents



#!/usr/bin/env python3

"""Main module for the README-AI CLI application."""

__package__ = "readmeai"

import asyncio
import os
import shutil
import tempfile
import traceback

from readmeai.cli.options import prompt_for_custom_image
from readmeai.config.settings import (
    AppConfig,
    AppConfigModel,
    ConfigHelper,
    GitSettings,
    ImageOptions,
    load_config,
    load_config_helper,
)
from readmeai.core.logger import Logger
from readmeai.core.model import ModelHandler
from readmeai.core.preprocess import RepoProcessor
from readmeai.markdown.headers import build_readme_md
from readmeai.markdown.tree import TreeGenerator
from readmeai.services.git_operations import clone_repo_to_temp_dir

logger = Logger(__name__)


async def readme_agent(conf: AppConfig, conf_helper: ConfigHelper) -> None:
    """Orchestrates the README file generation process."""
    llm = ModelHandler(conf)

    repo_url = conf.git.repository

    temp_dir = await asyncio.to_thread(tempfile.mkdtemp)
    try:
        await clone_repo_to_temp_dir(repo_url, temp_dir)

        tree_command(conf, conf_helper, temp_dir)

        parser = RepoProcessor(conf, conf_helper)
        dependencies, file_context = parser.get_dependencies(temp_dir)
        summaries = [(path, content) for path, content in file_context.items()]
        logger.info(f"Project dependencies: {dependencies}")
        logger.info(f"Project structure:\n{conf.md.tree}")

        async with llm.use_api() as api:
            if conf.cli.offline is False:
                prompts = [
                    {
                        "type": "summaries",
                        "context": {
                            "repo": repo_url,
                            "tree": conf.md.tree,
                            "dependencies": dependencies,
                            "summaries": summaries,
                        },
                    },
                    {
                        "type": "features",
                        "context": {
                            "repo": repo_url,
                            "tree": conf.md.tree,
                            "dependencies": dependencies,
                            "summaries": summaries,
                        },
                    },
                    {
                        "type": "overview",
                        "context": {
                            "name": conf.git.name,
                            "repo": repo_url,
                            "summaries": summaries,
                        },
                    },
                    {
                        "type": "slogan",
                        "context": {
                            "repo": repo_url,
                            "summaries": summaries,
                        },
                    },
                ]
                responses = await api.batch_text_generator(prompts)
                (
                    summaries_response,
                    features_response,
                    overview_response,
                    slogan_response,
                ) = (response for response in responses)
                summaries = summaries_response
                conf.md.features = conf.md.features.format(features_response)
                conf.md.overview = conf.md.overview.format(overview_response)
                conf.md.slogan = slogan_response
            else:
                (
                    summaries,
                    conf.md.features,
                    conf.md.overview,
                    conf.md.slogan,
                ) = (
                    [
                        (str(file_path), conf.md.default)
                        for file_path, _ in file_context.items()
                    ],
                    conf.md.features.format(conf.md.default),
                    conf.md.overview.format(conf.md.default),
                    conf.md.default,
                )

        build_readme_md(conf, conf_helper, dependencies, summaries)

    finally:
        shutil.rmtree(temp_dir)


def main(
    align: str,
    api_key: str,
    badges: str,
    emojis: bool,
    image: str,
    max_tokens: int,
    model: str,
    offline: bool,
    output: str,
    repository: str,
    temperature: float,
) -> None:
    """Main method of the readme-ai CLI application."""
    try:
        conf = load_config()
        conf_model = AppConfigModel(app=conf)
        conf_helper = load_config_helper(conf_model)
        conf = update_settings(
            conf,
            align,
            badges,
            emojis,
            image,
            max_tokens,
            model,
            offline,
            output,
            repository,
            temperature,
        )
        export_to_environment(conf, api_key)

        log_settings(conf)

        asyncio.run(readme_agent(conf, conf_helper))

    except Exception as exc_info:
        logger.error(
            f"Error generating README: {exc_info}\n{traceback.format_exc()}"
        )

    logger.info(
        "README-AI processing complete. File saved as - {conf.files.output}"
    )


def update_settings(
    conf: AppConfig,
    align: str,
    badges: str,
    emojis: bool,
    image: str,
    max_tokens: int,
    model: str,
    offline: bool,
    output: str,
    repository: str,
    temperature: float,
) -> AppConfig:
    """Sets up the application configuration."""
    conf.git = GitSettings(repository=repository)
    conf.files.output = output
    conf.cli.emojis = emojis
    conf.cli.offline = offline
    conf.llm.tokens_max = max_tokens
    conf.llm.model = model
    conf.llm.temperature = temperature
    conf.md.align = align
    conf.md.badges_style = badges
    conf.md.image = (
        image
        if image != ImageOptions.CUSTOM.name
        else prompt_for_custom_image(None, None, image)
    )
    return conf


def export_to_environment(config: AppConfig, api_key: str) -> None:
    """Set environment variables for the CLI application."""
    if api_key is not None:
        os.environ["OPENAI_API_KEY"] = api_key
    elif "OPENAI_API_KEY" not in os.environ:
        config.cli.offline = True


def tree_command(
    conf: AppConfig, conf_helper: ConfigHelper, temp_dir: str
) -> None:
    """Updates the markdown tree configuration."""
    tree_generator = TreeGenerator(
        conf_helper=conf_helper,
        root_dir=temp_dir,
        repo_url=conf.git.repository,
        repo_name=conf.git.name,
    )
    conf.md.tree = conf.md.tree.format(tree_generator.run())


def log_settings(conf: AppConfig) -> None:
    """Log the settings for the CLI application."""
    logger.info("Starting README-AI processing...")
    logger.info(f"Processing repository: {conf.git.repository}")
    logger.info(f"LLM model engine: {conf.llm.model}")
    logger.info(f"LLM temperature: {conf.llm.temperature}")
    logger.info(f"Badge style: {conf.md.badges_style}")
    logger.info(f"Image style: {conf.md.image}")
    logger.info(f"Header alignment: {conf.md.align}")
    logger.info(f"Using emojis: {conf.cli.emojis}")
    logger.info(f"Offline mode: {conf.cli.offline}")



"""Functions for building and formatting the badges in the README.md file."""

from readmeai.config.settings import AppConfig, GitService
from readmeai.core.factory import FileHandler
from readmeai.utils import utils


def badge_template(conf: AppConfig) -> str:
    """Return markdown template for badges"""
    if conf.git.source == GitService.LOCAL.value:
        return conf.md.badges_offline
    else:
        return conf.md.badges_shields


def build_html_badges(svg_icons: dict, dependencies: list) -> str:
    """Returns a list of badges for the project dependencies."""
    badges = [
        svg_icons[str(dependency).lower()]
        for dependency in dependencies
        if str(dependency).lower() in svg_icons
    ]
    # Sort badges by hex value (from light to dark color)
    badges.sort(key=lambda b: int(b[1], 16) if b[1] else 0, reverse=True)
    badges = [badge[0] for badge in badges]
    return format_html_badges(badges)


def format_html_badges(badges: list) -> str:
    """Formats the SVG icons into HTML <img src=""> tags."""
    badge_lines = []
    total_badges = len(badges)
    if total_badges < 8:
        badges_per_line = total_badges
    else:
        badges_per_line = total_badges // 2 + (total_badges % 2)

    if badges_per_line == 0:
        return ""

    for i in range(0, total_badges, badges_per_line):
        line = "\n".join(
            [
                f'<img src="{badge}" alt="{badge.split("/badge/")[1].split("-")[0]}">'
                for badge in badges[i : i + badges_per_line]
            ]
        )
        badge_lines.append(line)

    return "\n\n".join(badge_lines)


def shields_icons(conf: AppConfig, deps: list, full_name: str):
    """
    Generates badges for the README using shields icons, referencing the
    repository - https://github.com/Aveek-Saha/GitHub-Profile-Badges.
    """
    if conf.git.source == GitService.LOCAL.host:
        repo_host = GitService.LOCAL.host
    else:
        repo_host = GitService.extract_name_from_host(conf.git.source)

    resource_path = utils.get_resource_path(
        __package__, conf.files.shields_icons
    )

    md_badge_template = badge_template(conf)
    shields_dict = FileHandler().read(resource_path)
    shields_icons = md_badge_template.format(
        conf.md.align,
        build_html_badges(shields_dict, deps).format(conf.md.badges_style),
        repo_host,
        full_name,
        conf.md.badges_style,
    )
    return (
        utils.remove_substring(shields_icons)
        if "invalid" in full_name.lower()
        else shields_icons
    )


def skill_icons(conf: AppConfig, deps: list) -> str:
    """
    Generates badges for the README using skill icons, from the
    repository - https://github.com/tandpfun/skill-icons.
    """
    resource_path = utils.get_resource_path(
        __package__, conf.files.skill_icons
    )
    skill_icons_dict = FileHandler().read(resource_path)
    skill_icons = [
        icon for icon in skill_icons_dict["icons"]["names"] if icon in deps
    ]
    skill_icons.extend(["md", "github", "git"])
    icon_names = ",".join(skill_icons)
    # per_line = (len(skill_icons) + 2) // 2
    # icon_names = f"{icon_names}"  # &perline={per_line}"
    skill_icons = skill_icons_dict["url"]["base_url"] + icon_names

    if conf.md.badges_style == "skills-light":
        skill_icons = f"{skill_icons}&theme=light"

    return conf.md.badges_skills.format(
        alignment=conf.md.align,
        badges=skill_icons,
        image=conf.md.image,
        repo_name=conf.git.name.upper(),
        slogan=conf.md.slogan,
    )



"""Parser for gradle dependency files."""

import re
from typing import List

from readmeai.parsers.base_parser import FileParser

GRADLE_DECODE_ERROR = "Error decoding build.gradle content: {0}"


class BuildGradleParser(FileParser):
    """Parser for build.gradle dependency files."""

    def parse(self, content: str) -> List[str]:
        """Extracts package names from a build.gradle file."""
        try:
            pattern = r"(implementation|classpath|api|testImplementation|androidTestImplementation|kapt)\s+['\"]([^'\"]+)['\"]"
            matches = re.findall(pattern, content)
            package_names = set()
            for _, dependency in matches:
                parts = dependency.split(":")
                for part in parts:
                    part = part.split(".")
                    for p in part:
                        if p.isalpha():
                            package_names.add(p)

            return list(package_names)

        except re.error as error:
            self.log_error(GRADLE_DECODE_ERROR.format(error))
            return []


class BuildGradleKtsParser(FileParser):
    """Parser for build.gradle.kts dependency files."""

    def parse(self, content: str) -> List[str]:
        """Extracts package names from a build.gradle.kts file."""
        try:
            pattern = r"(\bimplementation|testImplementation)\s*\((['\"])([^'\"]+)\2\)"
            matches = re.findall(pattern, content)
            package_names = set()
            for _, _, dependency in matches:
                parts = dependency.split(":")
                for part in parts:
                    part = part.split(".")
                    for p in part:
                        if p.isalpha():
                            package_names.add(p)
            return list(package_names)

        except re.error as error:
            self.log_error(GRADLE_DECODE_ERROR.format(error))
            return []
